#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

import airflow.models.xcom
from airflow.providers.common.io.xcom.backend import XComObjectStorageBackend
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.utils import timezone

from tests_common.test_utils import db
from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, XCOM_RETURN_KEY

pytestmark = [pytest.mark.db_test]

if AIRFLOW_V_3_0_PLUS:
    from airflow.models.xcom import XComModel
    from airflow.sdk import ObjectStoragePath
    from airflow.sdk.execution_time.comms import XComResult
    from airflow.sdk.execution_time.xcom import resolve_xcom_backend
else:
    from airflow.io.path import ObjectStoragePath  # type: ignore[no-redef]
    from airflow.models.xcom import BaseXCom, resolve_xcom_backend  # type: ignore[no-redef]


@pytest.fixture(autouse=True)
def reset_db():
    """Reset XCom entries."""
    db.clear_db_runs()
    db.clear_db_xcom()
    yield
    db.clear_db_runs()
    db.clear_db_xcom()


@pytest.fixture(autouse=True)
def reset_cache():
    from airflow.providers.common.io.xcom import backend

    backend._get_base_path.cache_clear()
    backend._get_compression.cache_clear()
    backend._get_threshold.cache_clear()
    yield
    backend._get_base_path.cache_clear()
    backend._get_compression.cache_clear()
    backend._get_threshold.cache_clear()


@pytest.fixture
def task_instance(create_task_instance_of_operator, session):
    return create_task_instance_of_operator(
        EmptyOperator,
        dag_id="test-dag-id",
        task_id="test-task-id",
        logical_date=timezone.datetime(2021, 12, 3, 4, 56),
        session=session,
    )


class TestXComObjectStorageBackend:
    @pytest.fixture(autouse=True)
    def setup_test_cases(self, tmp_path):
        xcom_path = tmp_path / "xcom"
        xcom_path.mkdir()
        self.path = f"file://{xcom_path.as_posix()}"
        configuration = {
            ("core", "xcom_backend"): "airflow.providers.common.io.xcom.backend.XComObjectStorageBackend",
            ("common.io", "xcom_objectstorage_path"): self.path,
            ("common.io", "xcom_objectstorage_threshold"): "50",
        }
        with conf_vars(configuration):
            yield

    def test_value_db(self, task_instance, mock_supervisor_comms, session):
        session.add(task_instance)
        session.commit()
        XCom = resolve_xcom_backend()
        airflow.models.xcom.XCom = XCom

        XCom.set(
            key=XCOM_RETURN_KEY,
            value={"key": "value"},
            dag_id=task_instance.dag_id,
            task_id=task_instance.task_id,
            run_id=task_instance.run_id,
        )

        if AIRFLOW_V_3_0_PLUS:
            # When using XComObjectStorageBackend, the value is stored in the db is serialized with json dumps
            # so we need to mimic that same behavior below.
            mock_supervisor_comms.send.return_value = XComResult(key="return_value", value={"key": "value"})

        value = XCom.get_value(
            key=XCOM_RETURN_KEY,
            ti_key=task_instance.key,
        )
        assert value == {"key": "value"}

    def test_value_storage(self, task_instance, mock_supervisor_comms, session):
        session.add(task_instance)
        session.commit()
        XCom = resolve_xcom_backend()
        airflow.models.xcom.XCom = XCom

        XCom.set(
            key=XCOM_RETURN_KEY,
            value={"key": "bigvaluebigvaluebigvalue" * 100},
            dag_id=task_instance.dag_id,
            task_id=task_instance.task_id,
            run_id=task_instance.run_id,
        )

        if AIRFLOW_V_3_0_PLUS:
            XComModel.set(
                key=XCOM_RETURN_KEY,
                value=self.path,
                dag_id=task_instance.dag_id,
                task_id=task_instance.task_id,
                run_id=task_instance.run_id,
            )

            res = (
                XComModel.get_many(
                    key=XCOM_RETURN_KEY,
                    dag_ids=task_instance.dag_id,
                    task_ids=task_instance.task_id,
                    run_id=task_instance.run_id,
                    session=session,
                )
                .with_entities(XComModel.value)
                .first()
            )
            data = XComModel.deserialize_value(res)
        else:
            res = (
                XCom.get_many(
                    key=XCOM_RETURN_KEY,
                    dag_ids=task_instance.dag_id,
                    task_ids=task_instance.task_id,
                    run_id=task_instance.run_id,
                    session=session,
                )
                .with_entities(BaseXCom.value)
                .first()
            )
            data = BaseXCom.deserialize_value(res)

        p = XComObjectStorageBackend._get_full_path(data)
        assert p.exists() is True

        if AIRFLOW_V_3_0_PLUS:
            mock_supervisor_comms.send.return_value = XComResult(
                key=XCOM_RETURN_KEY, value={"key": "bigvaluebigvaluebigvalue" * 100}
            )

        value = XCom.get_value(
            key=XCOM_RETURN_KEY,
            ti_key=task_instance.key,
        )
        assert value == {"key": "bigvaluebigvaluebigvalue" * 100}

        if AIRFLOW_V_3_0_PLUS:
            qry = XComModel.get_many(
                key=XCOM_RETURN_KEY,
                dag_ids=task_instance.dag_id,
                task_ids=task_instance.task_id,
                run_id=task_instance.run_id,
                session=session,
            )
            assert str(p) == XComModel.deserialize_value(qry.first())
        else:
            qry = XCom.get_many(
                key=XCOM_RETURN_KEY,
                dag_ids=task_instance.dag_id,
                task_ids=task_instance.task_id,
                run_id=task_instance.run_id,
                session=session,
            )
            assert str(p) == qry.first().value

    def test_clear(self, task_instance, session, mock_supervisor_comms):
        session.add(task_instance)
        session.commit()
        XCom = resolve_xcom_backend()
        airflow.models.xcom.XCom = XCom

        XCom.set(
            key=XCOM_RETURN_KEY,
            value={"key": "superlargevalue" * 100},
            dag_id=task_instance.dag_id,
            task_id=task_instance.task_id,
            run_id=task_instance.run_id,
        )

        if AIRFLOW_V_3_0_PLUS:
            if hasattr(mock_supervisor_comms, "send_request"):
                # Back-compat of task-sdk. Only affects us when we manually create these objects in tests.
                last_call = mock_supervisor_comms.send_request.call_args_list[-1]
            else:
                last_call = mock_supervisor_comms.send.call_args_list[-1]
            path = (last_call.kwargs.get("msg") or last_call.args[0]).value
            XComModel.set(
                key=XCOM_RETURN_KEY,
                value=path,
                dag_id=task_instance.dag_id,
                task_id=task_instance.task_id,
                run_id=task_instance.run_id,
            )

            res = (
                XComModel.get_many(
                    key=XCOM_RETURN_KEY,
                    dag_ids=task_instance.dag_id,
                    task_ids=task_instance.task_id,
                    run_id=task_instance.run_id,
                    session=session,
                )
                .with_entities(XComModel.value)
                .first()
            )
            data = XComModel.deserialize_value(res)
        else:
            res = (
                XCom.get_many(
                    key=XCOM_RETURN_KEY,
                    dag_ids=task_instance.dag_id,
                    task_ids=task_instance.task_id,
                    run_id=task_instance.run_id,
                    session=session,
                )
                .with_entities(BaseXCom.value)
                .first()
            )
            data = BaseXCom.deserialize_value(res)
        p = XComObjectStorageBackend._get_full_path(data)
        assert p.exists() is True

        if AIRFLOW_V_3_0_PLUS:
            mock_supervisor_comms.send.return_value = XComResult(
                key=XCOM_RETURN_KEY, value={"key": "superlargevalue" * 100}
            )
        value = XCom.get_value(
            key=XCOM_RETURN_KEY,
            ti_key=task_instance.key,
        )
        assert value

        if AIRFLOW_V_3_0_PLUS:
            mock_supervisor_comms.send.return_value = XComResult(key=XCOM_RETURN_KEY, value=path)
            XCom.delete(
                dag_id=task_instance.dag_id,
                task_id=task_instance.task_id,
                run_id=task_instance.run_id,
                key=XCOM_RETURN_KEY,
                map_index=task_instance.map_index,
            )
            XComModel.clear(
                dag_id=task_instance.dag_id,
                task_id=task_instance.task_id,
                run_id=task_instance.run_id,
                map_index=task_instance.map_index,
            )
            value = (
                XComModel.get_many(
                    key=XCOM_RETURN_KEY,
                    dag_ids=task_instance.dag_id,
                    task_ids=task_instance.task_id,
                    run_id=task_instance.run_id,
                    session=session,
                )
                .with_entities(XComModel.value)
                .first()
            )
        else:
            XCom.clear(
                dag_id=task_instance.dag_id,
                task_id=task_instance.task_id,
                run_id=task_instance.run_id,
                session=session,
            )
            value = XCom.get_value(
                key=XCOM_RETURN_KEY,
                ti_key=task_instance.key,
                session=session,
            )

        assert p.exists() is False
        assert not value

    @conf_vars({("common.io", "xcom_objectstorage_compression"): "gzip"})
    def test_compression(self, task_instance, session, mock_supervisor_comms):
        session.add(task_instance)
        session.commit()

        XCom = resolve_xcom_backend()
        airflow.models.xcom.XCom = XCom

        XCom.set(
            key=XCOM_RETURN_KEY,
            value={"key": "superlargevalue" * 100},
            dag_id=task_instance.dag_id,
            task_id=task_instance.task_id,
            run_id=task_instance.run_id,
        )

        if AIRFLOW_V_3_0_PLUS:
            XComModel.set(
                key=XCOM_RETURN_KEY,
                value=self.path + ".gz",
                dag_id=task_instance.dag_id,
                task_id=task_instance.task_id,
                run_id=task_instance.run_id,
            )

            res = (
                XComModel.get_many(
                    key=XCOM_RETURN_KEY,
                    dag_ids=task_instance.dag_id,
                    task_ids=task_instance.task_id,
                    run_id=task_instance.run_id,
                    session=session,
                )
                .with_entities(XComModel.value)
                .first()
            )
            data = XComModel.deserialize_value(res)
        else:
            res = (
                XCom.get_many(
                    key=XCOM_RETURN_KEY,
                    dag_ids=task_instance.dag_id,
                    task_ids=task_instance.task_id,
                    run_id=task_instance.run_id,
                    session=session,
                )
                .with_entities(BaseXCom.value)
                .first()
            )
            data = BaseXCom.deserialize_value(res)

        assert data.endswith(".gz")

        if AIRFLOW_V_3_0_PLUS:
            mock_supervisor_comms.send.return_value = XComResult(
                key=XCOM_RETURN_KEY, value={"key": "superlargevalue" * 100}
            )

        value = XCom.get_value(
            key=XCOM_RETURN_KEY,
            ti_key=task_instance.key,
        )

        assert value == {"key": "superlargevalue" * 100}

    @pytest.mark.parametrize(
        "value, expected_value",
        [
            pytest.param(
                "file://airflow/xcoms/non_existing_file.json",
                "file://airflow/xcoms/non_existing_file.json",
                id="str",
            ),
            pytest.param(1, 1, id="int"),
            pytest.param(1.0, 1.0, id="float"),
            pytest.param("string", "string", id="str"),
            pytest.param(True, True, id="bool"),
            pytest.param({"key": "value"}, {"key": "value"}, id="dict"),
            pytest.param({"key": {"key": "value"}}, {"key": {"key": "value"}}, id="nested_dict"),
            pytest.param([1, 2, 3], [1, 2, 3], id="list"),
            pytest.param((1, 2, 3), (1, 2, 3), id="tuple"),
            pytest.param(None, None, id="none"),
        ],
    )
    def test_serialization_deserialization_basic(self, value, expected_value):
        def conditional_side_effect(data) -> ObjectStoragePath:
            if isinstance(data, str) and data.startswith("file://"):
                return ObjectStoragePath(data)
            return original_get_full_path(data)

        original_get_full_path = XComObjectStorageBackend._get_full_path

        with patch.object(
            XComObjectStorageBackend,
            "_get_full_path",
            side_effect=conditional_side_effect,
        ):
            XCom = resolve_xcom_backend()
            airflow.models.xcom.XCom = XCom

            serialized_data = XCom.serialize_value(value)
            mock_xcom_ser = MagicMock(value=serialized_data)
            deserialized_data = XCom.deserialize_value(mock_xcom_ser)

            assert deserialized_data == expected_value



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

from airflow.providers.common.compat.openlineage.facet import Dataset

from tests_common.test_utils.compat import ignore_provider_compatibility_error

with ignore_provider_compatibility_error("2.8.0", __file__):
    from airflow.providers.common.io.operators.file_transfer import FileTransferOperator


def test_file_transfer_copy():
    with mock.patch(
        "airflow.providers.common.io.operators.file_transfer.ObjectStoragePath"
    ) as mock_object_storage_path:
        source_path = mock.MagicMock()
        target_path = mock.MagicMock()
        mock_object_storage_path.side_effect = [source_path, target_path]
        source_path.exists.return_value = True
        target_path.exists.return_value = False
        operator = FileTransferOperator(
            task_id="test_common_io_file_transfer_task",
            src="test_source",
            dst="test_target",
        )
        operator.execute(context={})
        mock_object_storage_path.assert_has_calls(
            [
                mock.call("test_source", conn_id=None),
                mock.call("test_target", conn_id=None),
            ],
        )
        source_path.copy.assert_called_once_with(target_path)
        target_path.copy.assert_not_called()


def test_get_openlineage_facets_on_start():
    src_bucket = "src-bucket"
    src_key = "src-key"
    dst_bucket = "dst-bucket"
    dst_key = "dst-key"

    expected_input = Dataset(namespace=f"s3://{src_bucket}", name=src_key)
    expected_output = Dataset(namespace=f"s3://{dst_bucket}", name=dst_key)
    op = FileTransferOperator(
        task_id="test",
        src=f"s3://{src_bucket}/{src_key}",
        dst=f"s3://{dst_bucket}/{dst_key}",
    )

    lineage = op.get_openlineage_facets_on_start()
    assert len(lineage.inputs) == 1
    assert len(lineage.outputs) == 1
    assert lineage.inputs[0] == expected_input
    assert lineage.outputs[0] == expected_output


def test_get_openlineage_facets_on_start_without_namespace():
    mock_src = mock.MagicMock(key="/src_key", protocol="s3", bucket="src_bucket", sep="/")
    mock_dst = mock.MagicMock(key="dst_key", protocol="gcs", bucket="", sep="/")

    # Ensure the `namespace` attribute does not exist
    if hasattr(mock_src, "namespace"):
        delattr(mock_src, "namespace")
    if hasattr(mock_dst, "namespace"):
        delattr(mock_dst, "namespace")

    operator = FileTransferOperator(
        task_id="task",
        src=mock_src,
        dst=mock_dst,
        source_conn_id="source_conn_id",
        dest_conn_id="dest_conn_id",
    )
    # Make sure the _get_path method returns the mock objects
    operator._get_path = mock.Mock(side_effect=[mock_src, mock_dst])

    lineage = operator.get_openlineage_facets_on_start()
    assert len(lineage.inputs) == 1
    assert len(lineage.outputs) == 1
    assert lineage.inputs[0] == Dataset(namespace="s3://src_bucket", name="src_key")
    assert lineage.outputs[0] == Dataset(namespace="gcs", name="dst_key")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from pathlib import PosixPath
from urllib.parse import urlsplit, urlunsplit

import pytest

from airflow.providers.common.compat.openlineage.facet import Dataset as OpenLineageDataset
from airflow.providers.common.io.assets.file import (
    Asset,
    convert_asset_to_openlineage,
    create_asset,
    sanitize_uri,
)


@pytest.mark.parametrize(
    ("uri", "expected"),
    (
        ("file:///valid/path/", "file:///valid/path/"),
        ("file://C://dir/file", "file://C://dir/file"),
    ),
)
def test_sanitize_uri_valid(uri, expected):
    result = sanitize_uri(urlsplit(uri))
    assert urlunsplit(result) == expected


@pytest.mark.parametrize("uri", ("file://",))
def test_sanitize_uri_invalid(uri):
    with pytest.raises(ValueError):
        sanitize_uri(urlsplit(uri))


@pytest.mark.parametrize(
    ("path", "uri"),
    (
        ("/asdf/fdsa", "file:///asdf/fdsa"),
        ("file:///asdf/fdsa", "file:///asdf/fdsa"),
        ("file://asdf/fdsa", "file://asdf/fdsa"),
        ("file://127.0.0.1:8080/dir/file.csv", "file://127.0.0.1:8080/dir/file.csv"),
        (PosixPath("file:///tmpdir/some-path.csv"), "file:///tmpdir/some-path.csv"),
        ("file:/tmpdir/some-path.csv", "file:///tmpdir/some-path.csv"),
        ("file:tmpdir/some-path.csv", "file:///tmpdir/some-path.csv"),
    ),
)
def test_file_asset(path, uri):
    assert create_asset(path=path) == Asset(uri=uri)


@pytest.mark.parametrize(
    ("path", "ol_dataset"),
    (
        ("/valid/path", OpenLineageDataset(namespace="file", name="/valid/path")),
        (
            "file://127.0.0.1:8080/dir/file.csv",
            OpenLineageDataset(namespace="file://127.0.0.1:8080", name="/dir/file.csv"),
        ),
        ("file:///C://dir/file", OpenLineageDataset(namespace="file", name="/C://dir/file")),
        ("file://asdf.pdf", OpenLineageDataset(namespace="file", name="/asdf.pdf")),
        ("file:///asdf.pdf", OpenLineageDataset(namespace="file", name="/asdf.pdf")),
        (
            PosixPath("file:///tmp/pytest/test.log"),
            OpenLineageDataset(namespace="file", name="/tmp/pytest/test.log"),
        ),
        ("file:/tmp/pytest/test.log", OpenLineageDataset(namespace="file", name="/tmp/pytest/test.log")),
        ("file:tmp/pytest/test.log", OpenLineageDataset(namespace="file", name="/tmp/pytest/test.log")),
    ),
)
def test_convert_asset_to_openlineage(path, ol_dataset):
    result = convert_asset_to_openlineage(create_asset(path=path), None)
    assert result == ol_dataset



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
import uuid
from datetime import datetime
from typing import cast

from airflow import DAG

try:
    from airflow.sdk import task
except ImportError:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.providers.common.io.operators.file_transfer import FileTransferOperator
from airflow.utils.trigger_rule import TriggerRule

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import ObjectStoragePath
else:
    from airflow.io.path import ObjectStoragePath  # type: ignore[no-redef]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID")
DAG_ID = "example_file_transfer_local_to_s3"

SAMPLE_TEXT = "This is some sample text."

TEMP_FILE_PATH = ObjectStoragePath("file:///tmp")

AWS_BUCKET_NAME = f"bucket-aws-{DAG_ID}-{ENV_ID}".replace("_", "-")
AWS_BUCKET = ObjectStoragePath(f"s3://{AWS_BUCKET_NAME}")

AWS_FILE_PATH = AWS_BUCKET


@task
def create_temp_file() -> ObjectStoragePath:
    path = ObjectStoragePath(TEMP_FILE_PATH / str(uuid.uuid4()))

    with path.open("w") as file:
        file.write(SAMPLE_TEXT)

    return path


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_temp_file(path: ObjectStoragePath):
    path.unlink()


@task
def remove_bucket():
    AWS_BUCKET.rmdir(recursive=True)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),  # Override to match your needs
    tags=["example"],
    catchup=False,
) as dag:
    temp_file = create_temp_file()
    temp_file_path = cast("ObjectStoragePath", temp_file)

    # [START howto_transfer_local_to_s3]
    transfer = FileTransferOperator(src=temp_file_path, dst=AWS_BUCKET, task_id="transfer")
    # [END howto_transfer_local_to_s3]

    temp_file >> transfer >> remove_bucket() >> delete_temp_file(temp_file_path)

    from tests_common.test_utils.watcher import watcher

    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
